from __future__ import annotations

import datetime
import json
import os
import re
import string
import sys
import traceback
import types
import uuid
import warnings as pywarnings
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
import pytz
from IPython.core.getipython import get_ipython
from IPython.display import display, clear_output
from deprecated import deprecated

from seeq.base.seeq_names import SeeqNames
from seeq.sdk import *
from seeq.spy._datalab import is_jupyter, is_rkernel
from seeq.spy._errors import *

DEFAULT_DATASOURCE_NAME = 'Seeq Data Lab'
DEFAULT_DATASOURCE_CLASS = 'Seeq Data Lab'
DEFAULT_DATASOURCE_ID = 'Seeq Data Lab'
DEFAULT_DATASOURCE_DESCRIPTION = 'Signals, conditions and scalars from Seeq Data Lab.'

# This enables users to push to the Seeq-internal datasources. Pushing to the Metadata datasource isn't 100% accurate,
# but this should work for all item types.
INHERIT_FROM_WORKBOOK = SeeqNames.LocalDatasources.SeeqMetadataItemStorage.datasource_class

DEFAULT_WORKBOOK_PATH = 'Data Lab >> Data Lab Analysis'
DEFAULT_WORKBOOK_NAME = DEFAULT_WORKBOOK_PATH.split('>>')[-1].strip()
DEFAULT_WORKBOOK_STATE = '{"version":1, "state":{"stores":{}}}'
DEFAULT_WORKSHEET_NAME = 'From Data Lab'

DATALAB_HOME = '/home/datalab'

# Kept for backwards compatibility. Should match _folder#MY_FOLDER.
PATH_ROOT = '__My_Folder__'

DATASOURCE_MAP_ITEM_LEVEL_MAP_FILES = 'Item-Level Map Files'
DATASOURCE_MAP_REGEX_BASED_MAPS = 'RegEx-Based Maps'

EMPTY_GUID = '00000000-0000-0000-0000-000000000000'
GLOBALS_ONLY = EMPTY_GUID
GLOBALS_AND_ALL_WORKBOOKS = None

JOB_RESULTS_FOLDER_NAME = '_Job Results'
JOB_DATAFRAMES_FOLDER_NAME = '_Job DataFrames'


def present(row, column):
    return (column in row) and not (pd.api.types.is_scalar(row[column]) and pd.isnull(row[column]))


def get(row, column, default=None, assign_default=False):
    """
    Get the value in the column of a row. Can also accept dictionaries.
    Optionally define a default value/type to return and indicate if the
    default should be assigned to the row[column]. For example,

    >>> get(row, column, default=dict(), assign=False)
    if there is no value in row[column] a new dictionary will be returned

    >>> get(row, column, default=dict(), assign=True)
    if there is no value in row[column] a new dictionary will be assigned
    to row column and that dictionary will be returned

    :param row: The row object. Can be pd.Series, a single row pd.DataFrame, or dict
    :param column: The name of the column to query
    :param default: If the no value is present, the return value/type
    :param assign_default: Flag if the default should be assigned to the row[column]
    :return:
    """
    if present(row, column):
        return row[column]
    d = default
    if assign_default:
        row[column] = d
    return d


def get_timings(http_headers):
    output = dict()
    for header, cast in [('Server-Meters', int), ('Server-Timing', float)]:
        server_meters_string = http_headers[header]
        server_meters = server_meters_string.split(',')
        for server_meter_string in server_meters:
            server_meter = server_meter_string.split(';')
            if len(server_meter) < 3:
                continue

            dur_string = cast(server_meter[1].split('=')[1])
            desc_string = server_meter[2].split('=')[1].replace('"', '')

            output[desc_string] = dur_string

    return output


def raise_or_catalog(status, df=None, index=None, column=None, e=None, exception_type=None, message=None):
    if message is not None and exception_type is None:
        exception_type = SPyException
    if exception_type is not None:
        e = exception_type(message)
    if e is None:
        _, e, _ = sys.exc_info()
    if status is not None:
        if isinstance(e, KeyboardInterrupt):
            status.df['Result'] = 'Canceled'
        elif index is None:
            status.put('Result', format_exception(e))
        elif 'Result' in status.df.columns:
            status.df.at[index, 'Result'] = format_exception(e)
        status.update()
    if df is not None:
        df.at[index, column] = format_exception(e)
    if status.errors == 'raise':
        status.exception(e)
        raise e


def format_exception(e=None):
    exception_type = None
    tb = None
    if e is None:
        exception_type, e, tb = sys.exc_info()

    if isinstance(e, ApiException):
        return get_api_exception_message(e)

    else:
        if tb is not None:
            return '\n'.join(traceback.format_exception(exception_type, e, tb))
        else:
            return '[%s] %s' % (type(e).__name__, str(e))


GUID_REGEX = r'[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}'
GUID_REGEX_COMPILED = re.compile(GUID_REGEX, re.IGNORECASE)
HTML_EQUALS_REGEX = r'(?:=|&#61;)'
HTML_AMPERSAND_REGEX = r'(?:&(?!amp;)|&amp;)'
WORKSHEET_LINK_REGEX = fr'links\?type{HTML_EQUALS_REGEX}workstep{HTML_AMPERSAND_REGEX}' \
                       fr'workbook{HTML_EQUALS_REGEX}({GUID_REGEX}){HTML_AMPERSAND_REGEX}' \
                       fr'worksheet{HTML_EQUALS_REGEX}({GUID_REGEX}){HTML_AMPERSAND_REGEX}'
WORKSTEP_LINK_REGEX = fr'{WORKSHEET_LINK_REGEX}workstep{HTML_EQUALS_REGEX}({GUID_REGEX})'
EMAIL_REGEX = r"^([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~ \t]|(\\[\t -~]))+\")@([" \
              r"-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\[[\t -Z^-~]*])$"
UUID_LENGTH = 36


def is_guid(s: object):
    """
    Determine if an object is a string GUID/UUID

    Parameters
    ----------
    s : object
        The object to be tested

    Returns
    -------
    bool
        True if the object is a string GUID/UUID, False otherwise
    """
    return isinstance(s, str) and len(s) == UUID_LENGTH and GUID_REGEX_COMPILED.match(sanitize_guid(s)) is not None


def sanitize_guid(g):
    return g.upper()


def new_placeholder_guid():
    return str(uuid.uuid4()).upper()


def is_email(address: str) -> bool:
    return bool(re.match(EMAIL_REGEX, address))


def validate_timezone_arg(tz):
    if tz is not None:
        try:
            pd.to_datetime('2019-01-01T00:00:00.000Z').tz_convert(tz)

        except pytz.exceptions.UnknownTimeZoneError:
            raise SPyValueError('Unknown timezone "%s". Acceptable timezones:\n%s' %
                                (tz, '\n'.join(pytz.all_timezones)))


def validate_unique_dataframe_index(df: Optional[pd.DataFrame], var_name: str):
    if not isinstance(df, pd.DataFrame):
        return

    if not df.index.is_unique:
        raise SPyValueError(f"The {var_name} DataFrame's index must be unique. Use reset_index(drop=True, "
                            "inplace=True) before passing it in.")


def time_abbreviation_to_ms_multiplier(abbreviation):
    """
    Given a time abbreviation, returns a multiplier that converts the time
    value to ms and the canonical version of that time abbreviation.
    For example, "6 sec", pass the "sec" to this function and it will
    return (1000, 's'). Multiplying 6 by 1000 gives the duration in ms.

    Parameters
    ----------
    abbreviation : str
        The time abbreviation.  Acceptable values are:
        s, sec, second, seconds
        min, minute, minutes
        h, hr, hour, hours
        d, day, days
        w, wk, wks, week, weeks
        m, mo, month, months
        y, yr, yrs, year, years

    Returns
    -------
    (int, str)
        The multiplier to convert from the time basis to ms and the canonical
        version of the unit
    """
    abv = get_time_abbreviation(abbreviation)

    return get_abbreviation_multiplier_map()[abv], abv


def get_time_abbreviation(abbreviation):
    abv = abbreviation.lower()
    if abv in ['s', 'sec', 'second', 'seconds']:
        return 's'
    if abv in ['m', 'min', 'minute', 'minutes']:
        return 'min'
    if abv in ['h', 'hr', 'hour', 'hours']:
        return 'h'
    if abv in ['d', 'day', 'days']:
        return 'day'
    if abv in ['w', 'wk', 'wks', 'week', 'weeks']:
        return 'week'
    if abv in ['m', 'mo', 'month', 'months']:
        return 'month'
    if abv in ['y', 'yr', 'yrs', 'year', 'years']:
        return 'year'

    raise SPyValueError(f'Unrecognized time abbreviation "{abv}". Valid abbreviations: "sec", "min", "hr", "day", '
                        f'"wk", "mo", "yr"')


def get_abbreviation_multiplier_map():
    return {
        's': 1000,
        'min': 60000,
        'h': 3.6e6,
        'day': 8.64e7,
        'week': 6.048e8,
        'month': 2.628e9,
        'year': 3.154e10
    }


@deprecated(reason="Use parse_str_time instead")
def parse_str_time_to_ms(str_time):
    return parse_str_time(str_time)


def parse_str_time(str_time: str) -> Tuple[float, str, float]:
    """
    Given a string time like 5min, 3hr, 2d, 7mo, etc, get the numeric and unit
    portions and calculate the total number of milliseconds. Time units will be
    converted to the Seeq canonical versions, eg, 'sec' will yield 's'.
    For example '5min' returns (5, 'min', 300000)

    :param: str_time: A string representing a time. Eg, 3min
    :return: tuple (numeric portion, string unit, total milliseconds)
    """
    if not isinstance(str_time, str):
        raise SPyValueError(f'Duration string must be a string, instead got an {type(str_time).__name__}')

    match = re.match(r'([+-]?[\d.]+)\s*(\w+)', str_time)
    if not match:
        raise SPyValueError('Duration string not parseable: "%s"' % str_time)

    value = float(match.group(1))
    unit = match.group(2)
    multiplier, unit_name = time_abbreviation_to_ms_multiplier(unit)
    return value, unit_name, value * multiplier


def parse_str_time_to_timedelta(str_time):
    _, _, milliseconds = parse_str_time(str_time)
    return datetime.timedelta(milliseconds=milliseconds)


def test_and_set(target, target_key, source, source_key, apply_func=None, default_value=None, retain_target=True):
    """
    If the source has source_key and it's not NaN, set it as the value for target[target_key]

    Optionally, apply a function to the value in source[source_key]. Must be a function that
    takes a single input, such as lambda x: str.lower( x) or lambda x: str.upper(x) or

    Optionally, set a default input for target[target_key] if source[source_key] is NaN. Note that
    the default is only applied if source_key is in source, but the source value is NaN

    Parameters
    ----------
    target : dict
        The dictionary that the value is to be inserted into
    target_key : str
        The key in target for the value
    source : dict
        The source dictionary
    source_key : str
        The key in the source dict to look for the value
    apply_func : function
        A function to apply to the source[source_key] value, if it's not NaN
    default_value : obj
        If source[source_key] is NaN, the value for target[target_key]. See "retain_default" for
        additional functionality
    retain_target : bool
        If true and source_key not in source, do not assign the default - retain the current target value
        If false and source_key not in source, overwrite the current target value with the default
    """
    if source_key in source:
        if pd.notna(source[source_key]):
            if apply_func is not None:
                target[target_key] = next(map(apply_func, [source[source_key]]))
            else:
                target[target_key] = source[source_key]
            return
        elif default_value is not None:
            target[target_key] = default_value
        return
    elif not retain_target:
        target[target_key] = default_value


def validate_argument_types(expected_types):
    for _value, _name, _types in expected_types:
        if _value is None:
            continue

        if not isinstance(_value, _types):
            if isinstance(_types, tuple):
                acceptable_types = ' or '.join([_t.__name__ for _t in _types])
            else:
                acceptable_types = _types.__name__

            raise SPyTypeError("Argument '%s' should be type %s, but is type %s" % (_name, acceptable_types,
                                                                                    type(_value).__name__))

    return {_name: _value for _value, _name, _types in expected_types}


def none_to_nan(v):
    return np.nan if v is None else v


def ensure_unicode(o):
    if isinstance(o, bytes):
        return str(o, 'utf-8', errors='replace')
    else:
        return o


def force_to_data_frame(o):
    if isinstance(o, list):
        return pd.DataFrame(o)
    elif isinstance(o, dict):
        return pd.DataFrame([o])
    else:
        return o


def timer_start():
    return datetime.datetime.now()


def timer_elapsed(timer):
    return datetime.datetime.now() - timer


def convert_to_timestamp(value: Union[int, str], tz: Optional[str]) -> pd.Timestamp:
    """
    :param value: A date/time in either nanoseconds from the Unix Epoch or as ISO-8601 string (with time zone).
    :param tz: The time zone identifier to localize the value.
    :return: The value as a localized pd.Timestamp. pd.NA and None values return as NA.
    """
    if pd.isna(value):
        return pd.NA
    elif isinstance(value, str):
        value = pd.to_datetime(value)
    elif not isinstance(value, (int, float)):
        raise TypeError(f'Cannot convert {value} to timestamp. Expected either int or str.')
    return convert_timestamp_timezone(none_to_nan(pd.Timestamp(value)), tz)


def convert_timestamp_timezone(timestamp: pd.Timestamp, tz: Optional[str]) -> pd.Timestamp:
    if pd.isna(timestamp):
        return timestamp

    if timestamp.tzinfo is None:
        timestamp = timestamp.tz_localize('UTC')
    return timestamp.tz_convert(tz) if tz else timestamp


def print_output(output: str):
    if is_rkernel():
        import rpy2.robjects as robjects
        rcode = f"cat('{output}')"
        res = robjects.r(rcode)
        print(res)
    else:
        print(output)


def display_supports_html():
    if is_jupyter() or is_rkernel():
        return True
    else:
        # noinspection PyBroadException
        try:
            # noinspection PyUnresolvedReferences
            ipy_str = str(type(get_ipython()))
            if 'zmqshell' in ipy_str:
                return True
            if 'terminal' in ipy_str:
                return False

        except BaseException:
            return False


def ipython_clear_output(wait=False):
    clear_output(wait)


def ipython_display(*objs, include=None, exclude=None, metadata=None, transient=None, display_id=None, **kwargs):
    display(*objs, include=include, exclude=exclude, metadata=metadata, transient=transient,
            display_id=display_id, **kwargs)


def get_data_lab_datasource_input():
    datasource_input = DatasourceInputV1()
    datasource_input.name = DEFAULT_DATASOURCE_CLASS
    datasource_input.description = DEFAULT_DATASOURCE_DESCRIPTION
    datasource_input.datasource_class = DEFAULT_DATASOURCE_CLASS
    datasource_input.datasource_id = DEFAULT_DATASOURCE_ID
    datasource_input.stored_in_seeq = True
    datasource_input.additional_properties = [ScalarPropertyV1(name='Expect Duplicates During Indexing', value=True)]
    return datasource_input


def regex_from_query_fragment(query_fragment, contains=True):
    if query_fragment.startswith('/') and query_fragment.endswith('/'):
        regex = query_fragment[1:-1]
    else:
        regex = re.escape(query_fragment).replace(r'\?', '.').replace(r'\*', '.*')

        if contains and not regex.startswith('.*') and not regex.endswith('.*'):
            regex = '.*' + regex + '.*'

    return regex


CONSPICUOUS_TOKEN_REPRESENTING_A_SPACE = '>>>>>!SPACE!<<<<<'


def escape_regex(pattern):
    # re.escape() adds escapes to spaces, which makes for ugly Datasource Maps. We don't need spaces escaped because
    # we're not going to be dealing with verbose RegExes.
    # See https://stackoverflow.com/questions/32419837/why-re-escape-escapes-space
    pattern = pattern.replace(' ', CONSPICUOUS_TOKEN_REPRESENTING_A_SPACE)

    pattern = re.escape(pattern)

    pattern = pattern.replace(CONSPICUOUS_TOKEN_REPRESENTING_A_SPACE, ' ')

    return pattern


def get_signal_rollup_function_map():
    return {
        'average': 'average',
        'maximum': 'max',
        'minimum': 'min',
        'sum': 'sum'
    }


class RollUpFunction:
    def __init__(self, statistic, function, input_type, output_type, style):
        self.statistic = statistic
        self.function = function
        self.input_type = input_type
        self.output_type = output_type
        self.style = style

    def generate_formula(self, parameters: dict) -> str:
        param_list = list(parameters.keys())
        if len(parameters) > 1:
            if self.style == 'union':
                return ' or '.join(param_list)
            if self.style == 'intersect':
                return ' and '.join(param_list)
            if self.style == 'fluent':
                formula = param_list[0]
                for p in param_list[1:]:
                    formula += f'.{self.function}({p})'
                return formula
            return f'{self.function}({", ".join(param_list)})'
        elif len(parameters) == 1:
            parameter = param_list[0]
            if self.function.startswith('count'):
                return f'{self.function}({parameter})'
            if self.function == 'range':
                return '0.toSignal()'
            return parameter
        else:
            if self.output_type == 'Signal':
                return 'SCALAR.INVALID.toSignal()'
            if self.output_type == 'Condition':
                return "condition(40d, capsule('1970-01-01T00:00:00Z', '1970-01-01T00:00:00Z'))"
            return 'SCALAR.INVALID'


ROLL_UP_FUNCTIONS = [
    # Conditions
    RollUpFunction('union', 'union', 'Condition', 'Condition', 'union'),
    RollUpFunction('intersect', 'intersect', 'Condition', 'Condition', 'intersect'),
    RollUpFunction('counts', 'countOverlaps', 'Condition', 'Signal', 'vararg'),
    RollUpFunction('count overlaps', 'countOverlaps', 'Condition', 'Signal', 'vararg'),
    RollUpFunction('combine with', 'combineWith', 'Condition', 'Condition', 'vararg'),

    # Signals
    RollUpFunction('average', 'average', 'Signal', 'Signal', 'vararg'),
    RollUpFunction('maximum', 'max', 'Signal', 'Signal', 'fluent'),
    RollUpFunction('minimum', 'min', 'Signal', 'Signal', 'fluent'),
    RollUpFunction('range', 'range', 'Signal', 'Signal', 'vararg'),
    RollUpFunction('sum', 'add', 'Signal', 'Signal', 'vararg'),
    RollUpFunction('multiply', 'multiply', 'Signal', 'Signal', 'vararg'),
    RollUpFunction('combine with', 'combineWith', 'Signal', 'Signal', 'vararg'),

    # Scalars
    RollUpFunction('average', 'average', 'Scalar', 'Scalar', 'vararg'),
    RollUpFunction('maximum', 'max', 'Scalar', 'Scalar', 'fluent'),
    RollUpFunction('minimum', 'min', 'Scalar', 'Scalar', 'fluent'),
    RollUpFunction('sum', 'add', 'Scalar', 'Scalar', 'vararg'),
    RollUpFunction('multiply', 'multiply', 'Scalar', 'Scalar', 'vararg')
]


def get_signal_statistic_function_map():
    return {
        'average': 'average',
        'count': 'count',
        'delta': 'delta',
        'maximum': 'maxValue',
        'median': 'median',
        'minimum': 'minValue',
        'percentile': 'percentile',
        'range': 'range',
        'rate': 'rate',
        'standard deviation': 'stddev',
        'sum': 'sum',
        'totalized': 'totalized',
        'value at end': 'endValue',
        'value at start': 'startValue'
    }


def get_variable_number_statistic_function(statistic):
    match = re.match(r'^\w+\(([0-9]*.?[0-9]+|[0-9]+|[0-9]+.?[0-9]*)\)$', statistic)
    # matches numbers with decimal at beginning, end, or within
    if match is not None:
        # in Seeq, percentiles must be in the closed interval [0, 100].
        if 0 <= float(match.group(1)) <= 100:
            return statistic


def get_variable_time_unit_statistic_function(statistic):
    if re.match(r'^[\w ]+(\([\'"](s|min|h|day)[\'"]\))?$', statistic) is not None:
        return re.sub("'", '"', statistic)


def make_camel_case(string_, leading_upper=False):
    """
    Finds spaces or underscores and removes them, making the following letter
    upper case. If leading_upper=True, the first letter of the word will be
    made upper case too.

    leading_upper = False examples:
    "total duration" -> "totalDuration"
    "total_duration" -> "totalDuration"

    leading_upper = True examples:
    "total duration" -> "TotalDuration"
    "total_duration" -> "TotalDuration"
    """

    def camel(match):
        return match.group(1) + match.group(2).upper()

    snake_re = '(.*?)[_ ]([a-zA-Z])'
    new_string_ = re.sub(snake_re, camel, string_)
    return new_string_ if not leading_upper else new_string_[0].upper() + new_string_[1:]


def get_condition_statistic_function_map():
    return {
        'count ends': 'countEnds',
        'count starts': 'countStarts',
        'percent duration': 'percentDuration',
        'total duration': 'totalDuration'
    }


def statistic_to_aggregation_function(statistic, *, allow_condition_stats=True):
    statistic = statistic.lower()
    fns = get_signal_statistic_function_map()

    if allow_condition_stats:
        fns.update(get_condition_statistic_function_map())

    if 'percentile' in statistic and get_variable_number_statistic_function(statistic) is not None:
        return make_camel_case(get_variable_number_statistic_function(statistic))
    elif any([s in statistic for s in ['rate', 'total duration']]) and \
            get_variable_time_unit_statistic_function(statistic) is not None:
        return make_camel_case(get_variable_time_unit_statistic_function(statistic))
    elif statistic not in fns:
        raise SPyValueError('Statistic "%s" not recognized. Valid statistics:\n%s.\nNote that Rate can have '
                            'conversion keys of "s", "min", "h", or "day", eg, \'Rate("min")\', and Percentile '
                            'requires a number to specify the desired percentile, eg, "Percentile(25)"' %
                            (statistic.capitalize(), '\n'.join([string.capwords(s) for s in fns.keys()])))

    return fns[statistic] + '()'


def does_query_fragment_match(query_fragment, _string, contains=True):
    regex = regex_from_query_fragment(query_fragment, contains=contains)
    match = re.fullmatch(regex, _string, re.IGNORECASE)
    return match is not None


def get_workbook_type(arg):
    data: Optional[str] = None
    if isinstance(arg, WorkbookOutputV1):
        if arg.type in ['Analysis', 'Topic']:
            return arg.type

        if hasattr(arg, 'data'):
            data = arg.data
    else:
        data = arg

    if not data:
        return 'Analysis'

    if not isinstance(data, dict):
        # noinspection PyBroadException
        try:
            data = json.loads(data)
        except BaseException:
            return 'Analysis'

    if 'isReportBinder' in data and data['isReportBinder']:
        return 'Topic'
    else:
        return 'Analysis'


def safe_json_dumps(dictionary):
    if dictionary is None:
        return None

    def _nan_to_none(v):
        return None if isinstance(v, float) and np.isnan(v) else v

    def _nan_to_none_in_dict(d):
        d = d.copy()
        for k, v in d.items():
            if isinstance(v, dict):
                d[k] = _nan_to_none_in_dict(v)
            elif isinstance(v, list):
                d[k] = [_nan_to_none_in_dict(i) if isinstance(i, dict) else _nan_to_none(i) for i in v]
            else:
                d[k] = _nan_to_none(v)

        return d

    clean_dict = _nan_to_none_in_dict(dictionary)

    return json.dumps(clean_dict, skipkeys=True, allow_nan=False, indent=2, default=lambda o: '<not serializable>')


def path_string_to_list(path_string):
    return re.split(r'\s*>>\s*', path_string.strip())


def path_list_to_string(path_list):
    return ' >> '.join(path_list)


def sanitize_path_string(path):
    return path_list_to_string(path_string_to_list(path))


def string_to_formula_literal(s):
    if not isinstance(s, str):
        raise ValueError('Argument must be a string')

    s = re.sub(r"(['\\])", r"\\\1", s)
    return f"'{s}'"


def get_image_file(workbook_folder, image_id_tuple):
    return os.path.join(workbook_folder, 'Image_%s_%s' % image_id_tuple)


def save_image_files(image_dict: dict, folder: str):
    for image_id_tuple, content in image_dict.items():
        image_file = get_image_file(folder, image_id_tuple)
        with open(image_file, 'wb') as f:
            f.write(content)


def get_html_attr(fragment, attribute):
    attr_match = re.findall(r'\s+%s="(.*?)"' % attribute, fragment)
    return attr_match[0] if len(attr_match) > 0 else None


DATAFRAME_METADATA_CONTAINER_NAME = 'spy'


def put_properties_on_df(output_df, properties):
    """
    Adds the specified properties as metadata on output_df, under the "spy" attribute.

    Parameters
    ----------
    output_df : DataFrame
        DataFrame to which the properties will be attached. Typically, this is
        the output DataFrame of 'spy.search()' or 'spy.pull()'
    properties : types.SimpleNamespace
        Properties that will be added to the output_df.

    Returns
    -------
    {DataFrame}
        Original output_df, but now with the properties in "spy" attribute
    """
    if not isinstance(properties, types.SimpleNamespace):
        raise TypeError('properties argument must be a types.SimpleNamespace object')

    with pywarnings.catch_warnings():
        pywarnings.simplefilter('ignore', UserWarning)
        # Pandas issues a UserWarning in case the user is trying to add columns to the DataFrame as attributes which
        # is not allowed. We are not trying to add columns here though, it's a normal property

        setattr(output_df, DATAFRAME_METADATA_CONTAINER_NAME, properties)
        # noinspection PyProtectedMember
        if DATAFRAME_METADATA_CONTAINER_NAME not in output_df._metadata:
            # noinspection PyProtectedMember
            output_df._metadata.append(DATAFRAME_METADATA_CONTAINER_NAME)


def clear_properties_on_df(output_df):
    if hasattr(output_df, DATAFRAME_METADATA_CONTAINER_NAME):
        # noinspection PyBroadException
        try:
            delattr(output_df, DATAFRAME_METADATA_CONTAINER_NAME)
        except BaseException:
            pass


def look_up_in_df(item, lookup_df: pd.DataFrame) -> pd.DataFrame:
    clauses = list()
    friendly_strings = list()
    for col in ['Path', 'Asset', 'Name', 'Type']:
        col_value = get(item, col)

        if col_value is None:
            continue

        if col not in lookup_df.columns:
            raise SPyRuntimeError(f'Could not look up pushed item because lookup DataFrame does not have {col} column')

        if col == 'Type':
            col_value = col_value.replace('Stored', '').replace('Calculated', '').replace('Literal', '')
            clauses.append((lookup_df[col].str.contains(col_value)))
            friendly_strings.append(f'{col} contains "{col_value}"')
        else:
            clauses.append((lookup_df[col] == col_value))
            friendly_strings.append(f'{col} = "{col_value}"')

    friendly_string = '\n'.join(friendly_strings)
    if len(clauses) == 0:
        raise SPyRuntimeError(f'Could not find ID for pushed item due to empty search clause\n{friendly_string}')

    overall_clause = clauses[0]
    for clause in clauses[1:]:
        overall_clause &= clause

    pushed_item = lookup_df[overall_clause]
    if len(pushed_item) == 0:
        raise SPyRuntimeError(f'Could not find ID for pushed item where\n{friendly_string}')
    if len(pushed_item) > 1:
        raise SPyRuntimeError(f'Multiple IDs for pushed item where\n{friendly_string}')

    return pushed_item.iloc[0]


def ensure_upper_case_id(key, val):
    if key not in ['ID', 'Current Workstep ID', 'Asset ID', 'Report ID', 'Date Range ID', 'Asset Selection ID',
                   'Worksheet ID', 'Workstep ID', 'Condition ID', 'Item ID']:
        return val

    if not isinstance(val, str):
        return val

    result = ''
    last_index = 0
    for match in GUID_REGEX_COMPILED.finditer(val):
        result += val[last_index:match.start()]
        result += match.group(0).upper()
        last_index = match.end()

    result += val[last_index:]
    return result


def fix_up_ckeditor_curly_brace_weirdness(html):
    # Don't use re.sub()! For some reason there are weird edge cases where it doesn't work
    while True:
        matches = list(re.finditer(r'{\s*<!--\s*-->\s*{', html, re.MULTILINE))
        if not matches or len(matches) == 0:
            break

        match = matches[0]

        html = html[:match.start()] + '{{' + html[match.end():]

    return html


def get_shared_path_root(full_path_series):
    """
    Returns the highest shared name in the input paths. If no such name exists, returns None
    """
    first_full_path_list = full_path_series.iloc[0]
    if not len(first_full_path_list):
        return None
    root_name = first_full_path_list[0]
    all_roots_same = full_path_series.apply(lambda l: len(l) and l[0] == root_name).all()
    return root_name if all_roots_same else None


def get_attr(obj, *attr_list):
    """
    Retrieves an attribute from an object using a list of possibilities. Useful to handle compatibility situations
    where a dependency (like the generated Seeq SDK) has a renamed attribute.
    :param obj: The object in question
    :param attr_list: The set of attribute names to try, in the order desired
    :return: The attribute in question
    """
    for attr in attr_list:
        if hasattr(obj, attr):
            return getattr(obj, attr)

    raise AttributeError(f'{type(obj)} has no attributes {attr_list}')



from __future__ import annotations

import copy
import types
from typing import Optional, Union

import pandas as pd

from seeq import spy
from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.addons import _permissions


def search(query: Union[pd.DataFrame, pd.Series, dict, list, str], *, errors: Optional[str] = None,
           quiet: Optional[bool] = None, status: Status = None, session: Optional[Session] = None):
    """
    Issues a query to the Seeq Server to retrieve metadata for Add-on tools.
    This metadata can be used to update or uninstall Add-on tools.

    Parameters
    ----------
    query : {str, dict, list, pd.DataFrame, pd.Series}
        A mapping of property / match-criteria pairs

        If you supply a dict or list of dicts, then the matching
        operations are "contains" (instead of "equal to").

        If you supply a DataFrame or a Series, then the matching
        operations are "equal to" (instead of "contains").

        'Name' field allows you to query all installed Add-on Tools by using
        the wildcard '*'.

        Available options are:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        ID                  The ID of the Add-on tool
        Name                The name of the Add-on tool
        Description         The description of the Add-on tool
        Target URL          The URL that the Add-on opens
        Icon                Name of the fontawesome icon class displayed on the
                            Add-on tool
        Link Type           Display characteristics of the Add-on tool. Either
                            "window", "tab" or None
        Window Details      Display characteristics used when linkType is set
                            to "window"
        Sort Key            Determines the order in which the Add-on Tools are
                            displayed in the tool panel
        Reuse Window        True or False.
        =================== ===================================================

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame.

    quiet : bool, default False
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with rows for each item found and columns for each
        property.

        Additionally, the following properties are stored on the "spy"
        attribute of the output DataFrame:

        =================== ===================================================
        Property            Description
        =================== ===================================================
        func                A str value of 'spy.addons.search'
        kwargs              A dict with the values of the input parameters
                            passed to spy.addons.search to get the output
                            DataFrame
        status              A spy.Status object with the status of the
                            spy.addons.search call
        =================== ===================================================

    Examples
    --------
    Search for all the currently installed Add-on tools

    >>> search_results = spy.addons.search({'Name': '*'})

    To access the stored properties:
    >>> search_results.spy.kwargs
    >>> search_results.spy.status

    Search for an Add-on given its ID

    >>> search_results = spy.addons.search({"ID": "B540FF01-56B7-43AD-AE81-67AACB36C7F3"})

    Search for multiple Add-ons in the same spy.addons.search call:

    >>> search_results = spy.addons.search([{"Name": "Tool Name"}, {'Target URL': 'https://www.google.com'}])

    Search for multiple Add-ons using a pd.DataFrame

    >>> my_items = pd.DataFrame(
    >>>     {'Name': ['My Tool Name', 'Awesome Tool'],
    >>>      'Link Type': 'window'})
    >>> spy.addons.search(my_items)

    """

    input_args = _common.validate_argument_types([
        (query, 'query', (str, dict, list, pd.DataFrame, pd.Series)),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet, errors)
    session = Session.validate(session)
    _login.validate_login(session, status)

    try:
        return _search(session, query, status, input_args=input_args)

    except KeyboardInterrupt:
        status.update('Search canceled', Status.CANCELED)


def _search(session: Session, query, status: Status, *, input_args=None):
    system_api = SystemApi(session.client)
    tools = system_api.get_add_on_tools().add_on_tools
    if len(tools) == 0:
        status.update(
            f'There are zero Add-on tools installed',
            Status.SUCCESS)
        output_df = pd.DataFrame()
        add_properties_to_result(input_args, status, output_df)
        return output_df
    tool_dicts = [x.to_dict() for x in tools]
    for tool in tool_dicts:
        permissions = safely(lambda: _permissions.get_addon_permissions(tool['id'], session=session),
                             action_description=f"find permissions for addon with id {tool['id']}",
                             status=status,
                             default_value={"Groups": ["needs admin rights to display"],
                                            "Users": ["needs admin rights to display"]})
        tool['groups'] = permissions['Groups']
        tool['users'] = permissions['Users']

    tools_df = pd.DataFrame(tool_dicts)

    query_copy = copy.deepcopy(query)  # To avoid modifying the original object passed by the user

    if isinstance(query_copy, pd.DataFrame):
        queries = query_copy.to_dict(orient='records')
        comparison = '=='
    elif isinstance(query_copy, pd.Series):
        queries = [query_copy.to_dict()]
        comparison = '=='
    elif isinstance(query_copy, list):
        queries = query_copy
        comparison = '~='
    else:
        queries = [query_copy]
        comparison = '~='

    status.df = pd.DataFrame(queries)
    status.df['Time'] = 0
    status.df['Count'] = 0
    status.df['Result'] = 'Queued'
    status.update('Initializing', Status.RUNNING)

    results = []
    all_props = {'id': 'ID',
                 'name': 'Name',
                 'description': 'Description',
                 'target_url': 'Target URL',
                 'icon_class': 'Icon',
                 'link_type': 'Link Type',
                 'sort_key': 'Sort Key',
                 'reuse_window': 'Reuse Window',
                 'groups': "Groups",
                 'users': "Users",
                 'type': 'Type',
                 'is_archived': 'Archived',
                 'status_message': 'Status Message',
                 'window_details': 'Window Details'
                 }

    output_column_order = list(all_props.values())

    allowed_properties = ['ID', 'Name', 'Description', 'Target URL', 'Icon', 'Link Type', 'Sort Key',
                          'Reuse Window', 'Window Details']
    allowed_properties = {v: k for k, v in all_props.items() if v in allowed_properties}

    disallowed_properties = list()
    for status_index, current_query in enumerate(queries):
        for key, value in current_query.items():
            if key not in allowed_properties.keys():
                disallowed_properties.append(key)

        for key in disallowed_properties:
            if key in current_query:
                del current_query[key]

    if len(set(disallowed_properties)) > 0:
        disallowed_properties_str = '", "'.join(list(set(disallowed_properties)))
        message = f'The following properties are not indexed and will be ignored: "{disallowed_properties_str}"'
        status.warn(message)

    for status_index, current_query in enumerate(queries):
        timer = _common.timer_start()

        if not current_query:
            status.df.at[status_index, 'Time'] = _common.timer_elapsed(timer)
            status.df.at[status_index, 'Count'] = 0
            status.df.at[status_index, 'Result'] = 'Success'
            continue

        df_current_query = tools_df.copy()
        for prop in current_query.keys():
            # If ID is passed, then short-circuit the current query and return just the ID
            if 'ID' in current_query:
                query_str = f"id=='{current_query['ID']}'"
                df_current_query = tools_df.query(query_str)
                status.df.at[status_index, 'Count'] = len(df_current_query)
                break

            else:
                if prop == 'Name' and current_query['Name'] == "*":
                    # If "*" is passed, then keep the current list (all tools if no other props are passed)
                    status.df.at[status_index, 'Count'] = len(df_current_query)
                else:
                    if _common.present(current_query, prop):
                        query_str = _regex_from_query(current_query[prop].strip(), contains=(comparison == '~='))
                        df_current_query = df_current_query[
                            df_current_query[allowed_properties[prop]].str.contains(query_str, regex=True)]
                        status.df.at[status_index, 'Count'] = len(df_current_query)

        status.df.at[status_index, 'Time'] = _common.timer_elapsed(timer)
        status.df.at[status_index, 'Result'] = 'Success'
        results.extend(df_current_query.to_dict('records'))

    status.update('Query successful', Status.SUCCESS)
    output_df = pd.DataFrame(results)
    output_df.rename(columns=all_props, inplace=True)
    if not output_df.empty:
        output_df = output_df[output_column_order]
        duplicate_series = output_df['ID'].value_counts()
        number_of_duplicates = duplicate_series.sum() - len(duplicate_series)
        if number_of_duplicates > 0:
            output_df.drop_duplicates('ID', ignore_index=True, inplace=True)
            status.update(
                f'Query successful - Dropped {number_of_duplicates} duplicate{"s" if number_of_duplicates > 1 else ""}',
                Status.SUCCESS)

    add_properties_to_result(input_args, status, output_df)

    return output_df


def _regex_from_query(query_fragment, contains=True):
    regex = _common.regex_from_query_fragment(query_fragment, contains)
    if not contains:
        regex = f"^{regex}$"
    return regex


def add_properties_to_result(input_args, status, output_df):
    output_df_properties = types.SimpleNamespace(
        func='spy.addons.search',
        kwargs=input_args,
        status=status)

    _common.put_properties_on_df(output_df, output_df_properties)



from seeq.base import util
from seeq.spy import _datalab
from seeq.spy.jobs._pull import pull
from seeq.spy.jobs._push import push
from seeq.spy.jobs._schedule import schedule, unschedule


def get_notebook_url(*args, **kwargs):
    util.deprecation_warning('Use of spy.jobs.get_notebook_url() deprecated, use spy.utils.get_notebook_url() instead')
    return _datalab.get_notebook_url(*args, **kwargs)


__all__ = ['push', 'pull', 'schedule', 'unschedule', 'get_notebook_url']



import argparse
import datetime
import itertools
import os
import pickle
import time
from datetime import date
from hashlib import shake_128
from typing import List, Union, Dict, Optional

import pandas as pd
from numpy import nan
from seeq import spy
from seeq.base.seeq_names import SeeqNames
from seeq.spy._common import EMPTY_GUID
from seeq.spy.workbooks import Analysis, AnalysisWorksheet

AGILE_FILTER_NAME = ' Agile Filter'
AGILE_FILTER_FORMULA = '.agileFilter(1min)'
DELTA_NAME = ' 5min Delta'
DELTA_FORMULA = '.aggregate(delta(), periods(5min), middleKey()).max(0)'
PERCENTILE = 'P95'
ROOT_NODE = {'influxdb': 'Monitors-POC', 'legacy': 'Performance Monitors'}
DATASOURCES_WORKBOOK_SUFFIX = ' - Datasources'
DATASOURCE = 'Seeq Monitoring Script'

# On some versions of Seeq some monitor signals we use here are not available. This sets a tolerance for how
# many we can safely tolerate. If we have too many signals missing it's likely that something is wrong with our input
# arguments, and we raise an exception.
MAX_MISSING_SIGNALS = 10

# These are the columns that are returned by spy.push in the result dataframe when pushing signal metadata
PUSH_RESULT_COLUMNS = ['Datasource Class', 'Datasource ID', 'Data ID', 'ID', 'Push Result']

# These are hardcoded to our monitors site for simplicity, but can easily be externalized if needed
MONITORS_URL = "https://monitors.seeq.site/"
SEEQ_PERSONNEL_USER_GROUP_GUID = '792CFDF4-94B3-44B3-A2CA-7157230CC1E2'
CUSTOMER_MONITORS_FOLDER_ID = '54BAFDB1-E742-42C0-B912-8902281E722F'


def monitor_group_definitions(server: str, influxdb: bool):
    definitions = {
        'System Overview': [
            {'path': ['Cpu', 'System', 'UsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', 'Appserver', 'ServerLoadPercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Users', 'ActiveAuthenticatedUsers', 'Last15Min'],
             'name': 'Gauge'},
            {'path': ['Processes', 'Appserver', 'Uptime'],
             'name': 'Gauge'},
        ],
        'Cpu Overview': [
            {'path': ['Cpu', 'System', 'UsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Cpu', 'Appserver', 'UsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Appserver', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
        ],
        'Cpu Usage': [
            {'path': ['Processes', server, 'Appserver', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'JvmLink', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Nginx', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Postgres', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Renderer', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'ScalarCache', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'ScreenshotWorkers', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SeriesCache', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SeriesCachePostgres', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SupervisorCore', 'CpuUsagePercent'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
        ],
        'Memory overview': [
            {'path': ['Processes', server, 'Appserver', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'Appserver', 'AvailableMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'System', 'AvailablePhysicalMemoryMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Cache', 'NonSeries', 'Memory'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
        ],
        'Memory usage': [
            {'path': ['Processes', server, 'Appserver', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'JvmLink', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Nginx', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Postgres', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'Renderer', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'ScalarCache', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'ScreenshotWorkers', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SeriesCache', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SeriesCachePostgres', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
            {'path': ['Processes', server, 'SupervisorCore', 'MemoryUsageMB'],
             'name': 'Gauge',
             'calculated_signals': ['agile_filter']},
        ],
        'Garbage Collector': [
            {'path': ['Memory', 'GarbageCollector', 'Pause', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Memory', 'GarbageCollector', 'Pause', 'Timer'],
             'name': 'Mean',
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'GarbageCollector', 'Pause', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Memory', 'GarbageCollector', 'TotalPauseTime'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
        ],
        'Persistent Signal Cache': [
            {'path': ['Cache', 'Persistent', 'Signal', 'Read', 'Sample'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Cache', 'Persistent', 'Signal', 'Read', 'Time', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Cache', 'Persistent', 'Signal', 'Write', 'Time', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']}
        ],
        'Persistent Condition Cache': [
            {'path': ['Cache', 'Persistent', 'Condition', 'Read', 'Capsule'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Cache', 'Persistent', 'Condition', 'Read', 'Time', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Cache', 'Persistent', 'Condition', 'Write', 'Time', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']}
        ],
        'Disk Space': [
            {'path': ['HardDisk', 'Usage', 'AvailableBytes', 'DataFolder'],
             'name': 'Gauge'},
        ],
        'Disk MyData': [
            {'path': ['HardDisk', 'Usage', 'Folders', 'MyData', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'MyData', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'MyData', 'Size'],
             'name': 'Gauge'},
        ],
        'Disk Cache': [
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Content', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Series', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Thumbnail', 'FileCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Content', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Series', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Thumbnail', 'FolderCount'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Content', 'Size'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Series', 'Size'],
             'name': 'Gauge'},
            {'path': ['HardDisk', 'Usage', 'Folders', 'Cache', 'Thumbnail', 'Size'],
             'name': 'Gauge'},
        ],
        'Long running requests': [
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Submitted'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Duration', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued', 'Timer'],
             'name': 'P75',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Duration', 'Timer'],
             'name': 'P75',
             'calculated_signals': ['agile_filter']},
        ],
        'Users': [
            {'path': ['Users', 'ActiveAuthenticatedUsers', 'Last15Min'],
             'name': 'Gauge'},
            {'path': ['Users', 'BrowserTabs'],
             'name': 'Gauge'},
            {'path': ['Users', 'AuthenticatedUsers'],
             'name': 'Gauge'},
        ],
        'User jobs': [
            {'path': ['Threads', 'Jobs', 'User', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Queued', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Submitted'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Threads', 'Jobs', 'User', 'Duration', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
        ],
        'Date Range jobs': [
            {'path': ['Jobs', 'DateRangeJobs', 'Execution', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Jobs', 'DateRangeJobs', 'Execution', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Jobs', 'DateRangeJobs', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Jobs', 'DateRangeJobs', 'Misfired'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
        ],
        'Report jobs': [
            {'path': ['Jobs', 'ReportJobs', 'Execution', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Jobs', 'ReportJobs', 'Execution', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Jobs', 'ReportJobs', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Jobs', 'ReportJobs', 'Misfired'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
        ],
        'Async Migration jobs': [
            {'path': ['Jobs', 'AsyncMigrationStatus', 'Execution', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Jobs', 'AsyncMigrationStatus', 'Execution', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Jobs', 'AsyncMigrationStatus', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Jobs', 'AsyncMigrationStatus', 'Misfired'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
        ],
        'Thread Pool Queues': [
            {'path': ['Threads', 'HttpServer', 'Jobs', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'Request', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'LongRunning', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'Default', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'HttpServer', 'Initialization', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']}
        ],
        'Client Websocket': [
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Threads', 'ClientWebSocket', 'Messages', 'Duration', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']}
        ],
        'Agent Websocket': [
            {'path': ['Threads', 'AgentWebSocket', 'Messages', 'Running'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'AgentWebSocket', 'Messages', 'Queued'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Threads', 'AgentWebSocket', 'Messages', 'Completed'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Threads', 'AgentWebSocket', 'Messages', 'Duration', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']}
        ],
        'Network': [
            {'path': ['Network', 'PubSub', 'Channels', 'BroadcastChannel', 'Subscribers'],
             'name': 'Counter',
             'calculated_signals': ['agile_filter']},
            {'path': ['Network', 'AgentWebSocket', 'ConnectedAgents'],
             'name': 'Gauge'},
            {'path': ['Network', 'HttpServer', 'RestAPI', 'Samples'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Network', 'HttpServer', 'RestAPI', 'Capsules'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']}
        ],
        'Screenshots': [
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'Queued', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Timing', 'DataDownload', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Result', 'Aborted'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Result', 'Successful'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Queued'],
             'name': 'Gauge',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Screenshot', 'Running'],
             'name': 'Gauge',
             'calculated_signals': ['delta_agile_filter']}
        ],
        'Thumbnails': [
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': 'Count',
             'calculated_signals': ['delta_agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Timing', 'TotalRenderingTime', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Timing', 'Queued', 'Timer'],
             'name': PERCENTILE,
             'calculated_signals': ['agile_filter']},
            {'path': ['Rendering', 'HeadlessCapture', 'Thumbnail', 'Result', 'Aborted'],
             'name': 'Meter',
             'calculated_signals': ['delta_agile_filter']}
        ],
    }

    def map_name(v: str) -> str:
        if v == 'Gauge':
            return 'Gauge.value'
        if v == 'Counter' or v == 'Meter':
            return f"{v}.count"
        return v.lower()

    if influxdb:
        definitions = {key: [{k: map_name(v) if k == 'name' else v for k, v in d.items()} for d in value]
                       for key, value in definitions.items()}

    return definitions


EXCLUDED_DATASOURCES = {
    'Add-on Calculation Python UDFs',
    'Auth',
    'CassandraV2',
    'ConnectedCount',
    'CSV',
    'External Calculation Python UDFs',
    'FileFolders',
    'FileSignal',
    'Imported CSV Files',
    'Monitors',
    'OAuth 2_0',
    'Postgres Datums',
    'Rx',
    'Seeq - FileSignal',
    'Seeq Calculations',
    'Seeq Data Lab',
    'Seeq Metadata',
    'SeeqMonitors',
    'Seeq - Signal Datasource',
    'Time Series CSV Files',
}

COLOR_CODES = [
    '#0910a4', '#cfd1fc', '#c01311', '#f19695', '#8B6969', '#D2C3C3', '#d94bff', '#eca5ff', '#7e86d6', '#bec2ea',
    '#c70000', '#ff6363', '#0075d2', '#69bdff', '#e6d50a', '#f9f07e', '#ff5959', '#ffacac', '#ffd8ba', '#ffebdc',
    '#ff7d1a', '#ffbe8c', '#e3e751', '#f1f3a8', '#00ced1', '#b4feff', '#0a0dd8', '#7779f9', '#08d2fb', '#b5f1fe',
    '#C67171', '#E7C4C4', '#F4A460', '#FAD8BA', '#8B7355', '#D4C8B9', '#777733', '#D8D8A4', '#DB2645', '#F1A8B5',
    '#5C246E', '#CC97DD', '#23238E', '#9898E6', '#7e86d6', '#bec2ea', '#d94bff', '#eca5ff', '#c70000', '#ff6363',
]
MANY_COLOR_CODES = COLOR_CODES * 13


def assign_color_to_stored_signals(results: pd.DataFrame, lane: int, offset: int = 1) -> None:
    results['Color'] = MANY_COLOR_CODES[2 * lane - offset]


def customize_stored_signals_trend_display(results: pd.DataFrame, lane: int) -> pd.DataFrame:
    results['Line Width'] = 0.5
    results['Lane'] = lane
    results['Axis Group'] = lane,
    results['Line Style'] = 'Solid'
    assign_color_to_stored_signals(results, lane)
    return results


# Use this calculation to smooth a gauge or other counter that can go up or down such as CPU Usage or Job Run Time
def create_agile_filter_calculated_signal(name: str, formula_guid: str, lane: int) -> pd.DataFrame:
    signal_name = name.replace(' >> ', '.') + AGILE_FILTER_NAME + f'-{variable_digest(formula_guid, 3)}'
    df = pd.DataFrame([{
        'Name': signal_name,
        'Type': SeeqNames.Types.signal,
        'Formula': f'$s{AGILE_FILTER_FORMULA}',
        'Formula Parameters': {'$s': formula_guid},
        'Line Width': 1.5,
        'Lane': lane,
        'Axis Group': lane,
        'Color': MANY_COLOR_CODES[2 * lane - 2],
    }])
    df.set_index('Name', inplace=True, drop=False)
    return df


# Use this calculation for meters that only count upward such as those for Samples Read or Jobs Completed
def create_delta_agile_filter_calculated_signal(name: str, formula_guid: str, lane: int) -> pd.DataFrame:
    signal_name = name.replace(' >> ', '.') + DELTA_NAME + AGILE_FILTER_NAME + f'-{variable_digest(formula_guid, 3)}'
    df = pd.DataFrame([{
        'Name': signal_name,
        'Type': SeeqNames.Types.signal,
        'Formula': f'$s{DELTA_FORMULA}{AGILE_FILTER_FORMULA}',
        'Formula Parameters': {'$s': formula_guid},
        'Line Width': 1.5,
        'Lane': lane,
        'Color': MANY_COLOR_CODES[2 * lane - 2],
    }])
    df.set_index('Name', inplace=True, drop=False)
    return df


CALCULATED_SIGNAL_BUILDERS = {
    'agile_filter': create_agile_filter_calculated_signal,
    'delta_agile_filter': create_delta_agile_filter_calculated_signal,
}


def push_calculated_signals(calculated_signals_to_trend: pd.DataFrame, workbook_id: str,
                            worksheet_name: str) -> pd.DataFrame:
    start = time.time()
    print(f'Pushing calculated signals for worksheet {worksheet_name}')

    results = spy.push(
        metadata=calculated_signals_to_trend.reset_index(drop=True),
        datasource=DATASOURCE,
        workbook=workbook_id,
        worksheet=None,
        quiet=True
    )
    results.set_index('Name', inplace=True, drop=False)

    print(f'Finished pushing {len(results)} calculated signals for worksheet {worksheet_name} in'
          f' {int((time.time() - start) * 1000)} millis')

    return results


def init_sq_trend_store(stores):
    if 'sqTrendStore' not in stores:
        stores.update({'sqTrendStore': {'enabledColumns': {}}})


def enable_display_properties(stores) -> None:
    init_sq_trend_store(stores)
    stores['sqTrendStore']['enabledColumns']['SERIES'] = {"fullpath": True, "lane": True}


def enable_dimming(stores) -> None:
    init_sq_trend_store(stores)
    stores['sqTrendStore']['hideUnselectedItems'] = True


def set_trend_labels(label, location, stores) -> None:
    init_sq_trend_store(stores)
    stores['sqTrendStore']['labelDisplayConfiguration'] = {label: location}


def set_auto_update_display_range(stores) -> None:
    stores['sqDurationStore']['autoUpdate']['mode'] = 'MANUAL'
    stores['sqDurationStore']['autoUpdate']['manualInterval'] = {'value': 5, 'units': 'min', 'valid': True}


def set_worksheet_display_items(worksheet: AnalysisWorksheet, signals_to_trend: Union[List, pd.DataFrame]) -> None:
    if isinstance(signals_to_trend, List):
        signals_to_trend = pd.concat(signals_to_trend, ignore_index=True)
    worksheet.display_items = signals_to_trend
    stores = worksheet.current_workstep().get_workstep_stores()
    set_trend_labels('name', 'lane', stores)
    set_auto_update_display_range(stores)
    enable_display_properties(stores)
    enable_dimming(stores)


def create_worksheet_for_monitor(workbook: Analysis, monitor: str, start: datetime.date, end: datetime.date):
    worksheet = workbook.worksheet(monitor)
    worksheet.display_range = {
        'Start': start,
        'End': end
    }
    return worksheet


def get_stored_signal_lane_def(signal_metadata: pd.DataFrame, signal: Dict, lane: int, influxdb: bool) -> \
        pd.DataFrame:
    if influxdb:
        signal_name = signal['name'].split('.')
        signal_asset = signal_name[0] if len(signal_name) > 1 else signal['path'][-1]
        signal_path = build_path_str(signal['path'] if len(signal_name) > 1 else signal['path'][:-1])
        results = signal_metadata[(signal_metadata['Path'].str.lower().str.endswith(signal_path.lower())) &
                                  (signal_metadata['Asset'] == signal_asset) &
                                  (signal_metadata['Name'].str.lower() == signal_name[-1].lower())].copy()
        results['Name'] = build_name_str(signal['path'], signal_name[0])
    else:
        signal_name = build_name_str(signal['path'], signal['name'])
        results = signal_metadata[signal_metadata['Name'].str.lower() == signal_name.lower()].copy()

    if not results.empty:
        results['Lane'] = lane
        if 'calculated_signals' in signal:
            customize_stored_signals_trend_display(results, lane)
        else:
            assign_color_to_stored_signals(results, lane, 2)
    return results


def get_calculated_signal_lane_def(signal_metadata: pd.DataFrame, signal: dict, lane: int) -> Union[pd.DataFrame, None]:
    if 'calculated_signals' not in signal:
        return None
    if signal_metadata.empty:
        return None
    if pd.isnull(signal_metadata['ID'].item()):
        return None

    calculated_signals = [CALCULATED_SIGNAL_BUILDERS[filter_name] for filter_name in signal['calculated_signals']]
    path_in_name = signal['name'].split('.')
    signal_path = build_path_str(signal['path'], [path_in_name[0]]) if len(path_in_name) > 1 \
        else build_path_str(signal['path'])
    return pd.concat([
        filter_func(signal_path, stored_signal, lane)
        for stored_signal in signal_metadata['ID']
        for filter_func in calculated_signals
    ]).set_index('Name', drop=False)


def merge_push_results(lane_defs: List[pd.DataFrame], pushed_signals: pd.DataFrame) -> List[pd.DataFrame]:
    return [
        lane_def.merge(
            pushed_signals[PUSH_RESULT_COLUMNS],
            left_index=True,
            right_index=True
        )
        for lane_def in lane_defs
        if lane_def is not None
    ]


def create_regular_monitor_worksheets(workbook: Analysis,
                                      signal_metadata: pd.DataFrame,
                                      two_weeks_ago: datetime.date,
                                      today: datetime.date,
                                      server: str,
                                      influxdb: bool
                                      ) -> None:
    worksheet_definitions = {}
    calculated_signals_to_push = []
    missing = 0
    for monitor_group, monitor_definitions in monitor_group_definitions(server, influxdb).items():
        worksheet = create_worksheet_for_monitor(workbook, monitor_group, two_weeks_ago, today)
        stored_signal_lane_defs = [get_stored_signal_lane_def(signal_metadata, monitor_definition, i + 1, influxdb)
                                   for i, monitor_definition
                                   in enumerate(monitor_definitions)]
        for stored_signal_lane_def, monitor_definition in zip(stored_signal_lane_defs, monitor_definitions):
            if stored_signal_lane_def.empty:
                missing += 1
                name = build_name_str(monitor_definition['path'], monitor_definition['name'])
                print(f"A Stored Signal {name} wasn't found")

        if missing > MAX_MISSING_SIGNALS:
            raise RuntimeError(f"Too many missing Signals. Arguments to --server and --fqdn probably need to differ")

        calculated_signal_lane_defs = [
            get_calculated_signal_lane_def(stored_signal_lane_def, monitor_definition, i + 1)
            for i, (stored_signal_lane_def, monitor_definition)
            in enumerate(zip(stored_signal_lane_defs, monitor_definitions))
        ]
        calculated_signal_lane_defs = [s for s in calculated_signal_lane_defs if s is not None]

        worksheet_definitions[monitor_group] = {
            'worksheet': worksheet,
            'stored_signal_lane_defs': stored_signal_lane_defs,
            'calculated_signal_lane_defs': calculated_signal_lane_defs
        }

        if influxdb:
            for results in stored_signal_lane_defs:
                if not results.empty:
                    results['Formula'] = '$s'
                    results['Formula Parameters'] = [{'$s': results['ID'].iloc[0]}]
                    results['Type'] = 'CalculatedSignal'
                    results.set_index('Name', inplace=True, drop=False)
                    calculated_signals_to_push.append(results)
        calculated_signals_to_push.extend(calculated_signal_lane_defs)

    # remove ids for calculated signals constructed from stored signals
    if influxdb:
        for results in calculated_signals_to_push:
            if 'ID' in results.columns:
                results['Asset'] = nan
                results['Path'] = nan
                results.drop(columns=['ID', 'Datasource Name', 'Archived'], inplace=True)

    # some worksheets monitor the same signals, so we need to drop duplicates to successfully push everything
    signals_to_push = pd.concat(calculated_signals_to_push, ignore_index=True, copy=False
                                ).drop_duplicates(subset=['Name'])
    pushed_signals = push_calculated_signals(signals_to_push, workbook.id, '{All worksheets}')

    worksheet_definitions = {
        monitor_group: {
            **worksheet_definition,
            **({
                   'stored_signal_lane_defs': merge_push_results(
                       worksheet_definition['stored_signal_lane_defs'],
                       pushed_signals
                   )
               } if influxdb else {}),
            **{
                'calculated_signal_lane_defs': merge_push_results(
                    worksheet_definition['calculated_signal_lane_defs'],
                    pushed_signals
                )
            }
        }
        for monitor_group, worksheet_definition in worksheet_definitions.items()
    }
    for monitor_group, worksheet_definition in worksheet_definitions.items():
        worksheet = worksheet_definition['worksheet']
        lane_defs = [*worksheet_definition['stored_signal_lane_defs'],
                     *worksheet_definition['calculated_signal_lane_defs']]
        if len(lane_defs) > 0:
            set_worksheet_display_items(worksheet, lane_defs)


def create_datasource_monitor_worksheets(workbook: Analysis,
                                         signal_metadata: pd.DataFrame,
                                         two_weeks_ago: datetime.date,
                                         today: datetime.date,
                                         customer: str,
                                         fqdn: str,
                                         influxdb: bool) -> None:
    # Datasources are customer-dependent, so we have to go about them a little differently than other monitor data.
    base_path = f"{ROOT_NODE['influxdb' if influxdb else 'legacy']} >> {customer} >> {fqdn} >> Datasource"
    datasource_signal_paths = signal_metadata[signal_metadata['Path'].str.startswith(base_path)]['Path']
    datasources = datasource_signal_paths.str.split(' >> ', expand=True)
    if len(datasources.columns) <= 4:
        print(f'No Datasource paths found within path: {base_path}')
        return
    unique_datasources = set(sorted(datasources.iloc[:, 4].dropna()))
    eligible_datasources = sorted(unique_datasources - EXCLUDED_DATASOURCES)
    # It's more efficient to push all calcs in one call. Collect the worksheet->calcs, push all the calcs,
    # then add them into the worksheets with the desired formatting.
    worksheets_to_trend_items = []
    calculated_signals_to_push = []
    for datasource_class in eligible_datasources:
        datasource_class_path = f'{base_path} >> {datasource_class}'
        datasource_names = datasource_signal_paths[
            datasource_signal_paths.str.startswith(datasource_class_path)].str.split(' >> ', expand=True)
        unique_datasource_names = sorted(datasource_names.iloc[:, 5].dropna().unique()) \
            if len(datasource_names.columns) > 5 else []

        for datasource_name in unique_datasource_names:
            calculated_signals_to_trend = []
            signals_to_trend = []
            datasource_path = f'{datasource_class_path} >> {datasource_name}'

            # Timer Mean
            path = 'Rx >> Time'
            asset = 'Timer'
            name = 'mean' if influxdb else 'Mean'
            results = signal_metadata[(signal_metadata['Path'] == f'{datasource_path} >> {path}') &
                                      (signal_metadata['Asset'] == asset) &
                                      (signal_metadata['Name'].str.endswith(name))].copy()
            lane = 1
            if not results.empty:
                guid = results['ID'].iloc[0]
                calculated_signals_to_trend.append(create_agile_filter_calculated_signal(
                    f'{path}.{asset}.{name}', guid, 1))
                results['Name'] = f"{path.replace(' >>', '.')}.{asset}.{name}-{variable_digest(guid, 3)}"
                signals_to_trend.append(customize_stored_signals_trend_display(results, lane))
                lane += 1

            # Requests Gauge
            path = 'Requests' if influxdb else ''
            asset = 'Gauge' if influxdb else 'Requests'
            name = 'value' if influxdb else 'Gauge'
            results = signal_metadata[(signal_metadata['Path'] == f"{datasource_path}{' >> ' if path else ''}{path}") &
                                      (signal_metadata['Asset'] == asset) &
                                      (signal_metadata['Name'].str.endswith(name))].copy()
            if not results.empty:
                guid = results['ID'].iloc[0]
                results['Lane'] = lane
                results['Axis Group'] = lane,
                results['Name'] = f"{path.replace(' >>', '.')}.{asset}.{name}-{variable_digest(guid, 3)}"
                signals_to_trend.append(results)
                lane += 1

            # Receive Meters
            for monitor in ['Messages', 'Samples', 'Successes', 'Failures']:
                path = f'Rx >> {monitor}' if influxdb else 'Rx'
                asset = 'Meter' if influxdb else monitor
                name = 'count' if influxdb else 'Meter'
                results = signal_metadata[(signal_metadata['Path'].str.startswith(f'{datasource_path} >> {path}')) &
                                          (signal_metadata['Asset'] == asset) &
                                          (signal_metadata['Name'].str.endswith(name))].copy()
                if not results.empty:
                    guid = results['ID'].iloc[0]
                    calculated_signals_to_trend.append(create_delta_agile_filter_calculated_signal(
                        f'{path}.{asset}.{name}', guid, lane))
                    results['Name'] = f"{path.replace(' >>', '.')}.{asset}.{name}-{variable_digest(guid, 3)}"
                    signals_to_trend.append(customize_stored_signals_trend_display(results, lane))
                    lane += 1

            if len(calculated_signals_to_trend) > 0:
                worksheet = workbook.worksheet(f'{datasource_class} >> {datasource_name}'[:99])
                worksheet.display_range = {'Start': two_weeks_ago, 'End': today}
                worksheets_to_trend_items.append({
                    'Worksheet': worksheet,
                    'Stored Signals': signals_to_trend,
                    'Calculated Signals': calculated_signals_to_trend,
                })
                if influxdb:
                    for results in signals_to_trend:
                        if not results.empty:
                            results['Asset'] = nan
                            results['Path'] = nan
                            results['Formula'] = '$s'
                            results['Formula Parameters'] = [{'$s': results['ID'].iloc[0]}]
                            results['Type'] = 'CalculatedSignal'
                            results.set_index('Name', inplace=True, drop=False)
                            results.drop(columns=['ID', 'Datasource Name', 'Archived'], inplace=True)
                            calculated_signals_to_push.append(results)
                calculated_signals_to_push.extend(calculated_signals_to_trend)

    signals_to_push = pd.concat(calculated_signals_to_push, ignore_index=True, copy=False).drop_duplicates(
        subset=['Name'])
    pushed_signals = push_calculated_signals(signals_to_push, workbook.id, '{All datasource worksheets}')
    for ws_and_items in worksheets_to_trend_items:
        worksheet = ws_and_items['Worksheet']
        stored_signals = merge_push_results(ws_and_items['Stored Signals'], pushed_signals)
        calculated_signals = merge_push_results(ws_and_items['Calculated Signals'], pushed_signals)

        all_signals = pd.concat(stored_signals + calculated_signals, ignore_index=True)
        all_signals['Type'] = 'Signal'
        set_worksheet_display_items(worksheet, all_signals)


def push_workbook(workbook: Analysis,
                  workbook_name: str) -> None:
    print(f'Creating Workbook {workbook_name}.')
    start = time.time()
    url = spy.workbooks.push(
        workbook,
        path=CUSTOMER_MONITORS_FOLDER_ID,
        datasource=DATASOURCE,
        refresh=False,
        quiet=True
    )['URL'][0]
    print(f'Finished creating workbook {workbook_name} in {int((time.time() - start) * 1000)} millis.\n'
          f'Go to {url} to access the workbook.')
    spy.acl.push(workbook.id, {
        'ID': SEEQ_PERSONNEL_USER_GROUP_GUID,
        'Read': True,
        'Write': True,
        'Manage': True,
    }, replace=True, disable_inheritance=False, quiet=True)
    print(f'Gave access for {workbook_name} to Seeq Personnel.')


def build_path_str(*path_lists):
    return ' >> '.join(itertools.chain(*path_lists))


def build_name_str(path, name):
    return '.'.join([*path, name])


def variable_digest(data: str, length: int):
    h = shake_128()
    h.update(data.encode())
    return h.hexdigest(length)


def storage_load(key: str, expiration=24 * 60 * 60):
    try:
        file = f".{key}.pickle"
        last_modified = os.path.getmtime(file)
        if time.time() - expiration <= last_modified:
            with open(file, 'rb') as handle:
                return pickle.load(handle)
        return None
    except:
        return None


def storage_dump(key: str, data):
    try:
        with open(f".{key}.pickle", 'wb') as handle:
            return pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
    except:
        pass


def get_asset_tree_signals(session: spy.Session, path: str, root_id: Optional[str] = None,
                           use_cache: bool = True) -> pd.DataFrame:
    print(f"Started retrieving Signals in Asset Tree: {path}")
    start = time.time()
    file_name = f"tree_{path.replace('>', '_').replace(' ', '')}"
    tree = storage_load(file_name) if use_cache else None
    if tree is not None:
        print(f"Finished retrieving {len(tree)} Signals from Asset Tree cache {file_name}")
        return tree
    # Calling SPy Search with only Path (no Type or other options), means it'll use the TreesApi traversal.
    # This is more efficient for this case since we know the tree is limited in scope and mostly contains signals.
    # As a bonus, it also avoids the CRAB-33245 bug.
    tree = spy.search({'Asset': root_id} if root_id else {'Path': path},
                      workbook=EMPTY_GUID,
                      session=session,
                      errors='catalog',
                      quiet=True,
                      recursive=True,
                      all_properties=False)
    if tree.empty:
        raise RuntimeError(f"Couldn't find a tree for {path}. Arguments --customer and/or --fqdn are likely incorrect.")

    tree = tree[tree['Type'].str.contains("Signal") & tree['Path'].str.contains(path)]
    storage_dump(file_name, tree)
    print(f"Finished retrieving {len(tree)} Signals from Asset Tree in {int((time.time() - start) * 1000)} millis")
    return tree


def main(
        url: str,
        user: str,
        access_key: str,
        password: str,
        customer: str,
        server: str,
        fqdn: str,
        workbook_name: str,
        overwrite: bool,
        monitoring: str,
        root_id: Optional[str] = None,
        use_cache: bool = True,
        influxdb: bool = True,
) -> None:
    if monitoring not in ['sh', 'ds', 'both']:
        raise 'Provide one of the following for -m flag "ds" for datasources, "sh" for system health or "both"'

    start = time.time()
    spy.options.allow_version_mismatch = True
    spy.login(url=url, access_key=access_key, username=user, password=password, ignore_ssl_errors=True, quiet=True)
    print("Logged in successfully")
    today = date.today()
    fqdn = fqdn if fqdn else server
    two_weeks_ago = today - datetime.timedelta(days=14)
    workbook_name = workbook_name if workbook_name else f'Monitoring {customer} - {fqdn}'
    worksheet_name = list(monitor_group_definitions(server, influxdb).keys())[0]

    subroot_path = f"{ROOT_NODE['influxdb' if influxdb else 'legacy']} >> {customer} >> {fqdn}"
    signal_metadata = get_asset_tree_signals(spy.session, subroot_path, root_id, use_cache)

    if monitoring == 'sh' or monitoring == 'both':
        workbook_df = spy.search({'Name': workbook_name, 'Type': 'Workbook'}, quiet=True)
        workbook_df.sort_values('ID')
        if not workbook_df.empty and not overwrite:
            print(
                f'{workbook_name} already exists and -o/--overwrite is false. Please set -o/--overwrite to true if you '
                f'would like to overwrite the existing monitoring workbook.')
            return

        if len(workbook_df) > 0:
            workbook = spy.workbooks.Analysis({'Name': workbook_name, 'ID': workbook_df['ID'][0]})
        else:
            workbook = spy.workbooks.Analysis({'Name': workbook_name})
            workbook.worksheet(worksheet_name)
            spy.workbooks.push(workbook, quiet=True)

        try:
            create_regular_monitor_worksheets(workbook, signal_metadata, two_weeks_ago, today, server, influxdb)
        except Exception as e:
            raise Exception(f'Encountered error while building workbook {workbook_name}. No workbook was built') from e
        push_workbook(workbook, workbook_name)

    if monitoring == 'ds' or monitoring == 'both':
        datasource_workbook_name = workbook_name + DATASOURCES_WORKBOOK_SUFFIX
        workbook_df = spy.search({'Name': datasource_workbook_name, 'Type': 'Workbook'}, quiet=True)
        workbook_df.sort_values('ID')
        if len(workbook_df) > 0:
            workbook_datasources = spy.workbooks.Analysis(
                {'Name': datasource_workbook_name, 'ID': workbook_df['ID'][0]})
        else:
            workbook_datasources = spy.workbooks.Analysis({'Name': datasource_workbook_name})
            workbook_datasources.worksheet(worksheet_name)
            spy.workbooks.push(workbook_datasources, quiet=True)

        try:
            create_datasource_monitor_worksheets(workbook_datasources, signal_metadata, two_weeks_ago, today,
                                                 customer, fqdn, influxdb)
        except BaseException as e:
            raise Exception(f'Encountered error while building workbook {datasource_workbook_name}.\n'
                            f'No workbook was built.') from e
        push_workbook(workbook_datasources, datasource_workbook_name)
    print(f'Script completed in {int((time.time() - start) * 1000)} millis.')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-u", "--user", required=False, help="User name to log in with")
    parser.add_argument("-p", "--password", required=False, help="Password name to log in with")
    parser.add_argument("-k", "--access_key", required=False, help="Access key to log in with")
    parser.add_argument("-c", "--customer", required=True, help="Customer name for which to build monitoring workbook")
    parser.add_argument("-s", "--server", required=True, help="Customer server name for which to build monitoring "
                                                              "workbook (e.g. xyz-main-p) - case-sensitive")
    parser.add_argument("-f", "--fqdn", required=False,
                        help="Specify a fully-qualified domain name (e.g. xyz.seeq.site) if different from server "
                             "name (xyz-main-p) - case-insensitive, defaults to -s|--server")
    parser.add_argument("-w", "--workbook_name", required=False, help="The name for the monitoring workbook")
    parser.add_argument("-o", "--overwrite", required=False, action='store_true', default=True,
                        help="Overwrite the monitoring workbook if the name already exists")
    parser.add_argument("-m", "--monitoring", default="both", required=False,
                        choices=['sh', 'ds', 'both'],
                        help="Create system health (\"sh\"), datasource (\"ds\") or workbooks monitoring both ("
                             "\"both\").")
    parser.add_argument("-r", "--root_guid", required=False,
                        help="GUID of the Customer Root Asset - useful for cases when there are more than one folder "
                             "with the same name (due to a bad configuration or errors on the monitor ingestion "
                             "proces).")
    parser.add_argument("-n", "--no_cache", required=False, action='store_true',
                        help="Use local cache for the asset tree")
    parser.add_argument("-i", "--influxdb", required=False, action='store_true', default=True,
                        help="Use the new InfluxDB monitors or the legacy File-based monitors")

    args = parser.parse_args()
    main(MONITORS_URL, args.user, args.access_key, args.password, args.customer, args.server, args.fqdn,
         args.workbook_name, args.overwrite, args.monitoring, args.root_guid, not args.no_cache, args.influxdb)


def test_build_path_str():
    assert build_path_str(
        ['Path component 1', 'two'], ['More', 'paths', 'Here'], ['and']
    ) == 'Path component 1 >> two >> More >> paths >> Here >> and'


def test_build_name_str():
    assert build_name_str(
        ['Path component 1', 'two'], 'three'
    ) == 'Path component 1.two.three'



import argparse
from concurrent import futures as cf
from concurrent.futures import Executor
from concurrent.futures import Future
from datetime import datetime
from typing import List, Dict, Optional, Any, Callable, Tuple

from seeq import spy, sdk
from seeq.sdk import ApiClient, ConditionUpdateInputV1, ScalarPropertyV1, ConditionMonitorInputV1, PropertyInputV1, \
    ConditionBatchInputV1

formula = "($s.setRefreshRate(1min){threshold})\n" \
          ".removeShorterThan(2*{ignore_gaps})\n" \
          ".merge({ignore_gaps})\n" \
          ".removeShorterThan({min_duration})\n" \
          ".setMaximumDuration(1day)\n" \
          ".setProperty('Customer','{customer}')\n" \
          ".setProperty('Domain','{domain}')\n" \
          ".setProperty('Server','{server}')\n" \
          ".setProperty('Metric', $s, average())\n" \
          ".setProperty('Workbook', '{workbook_url}')"
formula_bindings = {
    'Cpu.System.UsagePercent.Gauge': {'threshold': '>=0.9', 'min_duration': '5min', 'ignore_gaps': '1min'},
    'Processes.Appserver.ServerLoadPercent.Gauge': {'threshold': '>=2', 'min_duration': '5min',
                                                    'ignore_gaps': '1min'},
    'Threads.HttpServer.LongRunning.Queued.Timer.p75': {'threshold': '>10000', 'min_duration': '5min',
                                                        'ignore_gaps': '1min'},
}

DATASOURCE = 'Seeq Monitoring Script'

# This is hardcoded to our monitors site for simplicity, but can easily be externalized if needed
MONITORS_URL = "https://monitors.seeq.dev/"


def get_session_info(session: spy.Session) -> Tuple[str, str, str]:
    return session.client.host, session.client.auth_token, session.client.csrf_token  # type: ignore


def get_api_client(session_info: Tuple[str, str, str], disable_retries: bool = False) -> ApiClient:
    client = ApiClient(host=session_info[0])
    client.auth_token = session_info[1]
    client.csrf_token = session_info[2]
    if disable_retries:
        client.configuration.retry_timeout_in_seconds = 0
    return client


def run_parallel(job_name: str, fns: List[Callable], fn_args: List = (), use_threads: bool = True,
                 max_parallelism: Optional[int] = None, executor: Optional[Executor] = None,
                 timeout: Optional[int] = None) -> Tuple[List, List[Optional[Exception]], bool]:
    # Initialize result and error lists
    results = [False] * len(fns)
    errors: List[Optional[Exception]] = [None] * len(fns)

    # Run in same process if only one function is given or max_parallelism is 1
    if len(fns) <= 1 or max_parallelism == 1:
        has_errors = False
        for index, fn in enumerate(fns):
            try:
                results[index] = fn(*fn_args[index])
            except Exception as e:
                errors[index] = e
                has_errors = True
        return results, errors, has_errors

    # Execute in parallel and wait for all futures to finish
    local_executor = None
    try:
        local_executor = executor if executor is not None else \
            cf.ThreadPoolExecutor(max_workers=max_parallelism,
                                  thread_name_prefix=job_name) if use_threads else cf.ProcessPoolExecutor(
                max_workers=max_parallelism)
        futures: Dict[int, Future] = {
            index: (local_executor.submit(fn) if fn_args is None else local_executor.submit(fn, *fn_args[index]))
            for index, fn in enumerate(fns)
        }
        cf.wait(futures.values(), timeout=timeout)
    finally:
        if executor is None and local_executor is not None:
            local_executor.shutdown()

    # Collect results and errors
    for index, future in futures.items():
        try:
            results[index] = future.result(timeout=0)
        except Exception as e:
            errors[index] = e

    has_errors = len(list(filter(lambda entry: entry is not None, errors))) > 0

    return results, errors, has_errors


def now_suffix() -> str:
    return f"{datetime.now().isoformat(timespec='seconds').replace(':', '')}Z"


def delete_items(client, items: List[Dict], delete_also: bool = False):
    client.configuration.retry_timeout_in_seconds = 1
    api = sdk.ItemsApi(client)
    suffix = f"_DELETED_{now_suffix()}"
    for item in items:
        id = item['id']

        api.archive_item(id=id, delete=False, archived_reason="BY_USER")
        if delete_also:
            try:
                api.archive_item(id=id, delete=True, archived_reason="BY_USER")
            except Exception as e:
                if 'Data ID' in item:
                    new_data_id = f"{item['Data ID']}{suffix}"
                    api.set_property(id=id, property_name='Data ID', body=PropertyInputV1(value=new_data_id))


def assign_properties(item: Dict[str, Any], properties: List[Dict[str, Any]]) -> Dict[str, Any]:
    for p in properties:
        name = p['name'][0].lower() + p['name'][1:].replace(' ', '')
        item[name] = p['value'] + ('' if p.get('unitOfMeasure', 'string') == 'string' else
                                   p['unitOfMeasure'])
    return item


def get_item(session_info: Tuple[str, str, str], guid: str) -> Dict[str, Any]:
    return sdk.ItemsApi(api_client=get_api_client(session_info)).get_item_and_all_properties(
        id=guid, _response_type='json'  # type: ignore
    )


def get_item_properties(session_info: Tuple[str, str, str],
                        item_ids: List[str],
                        max_parallelism: Optional[int] = None,
                        executor: Optional[Executor] = None,
                        timeout: Optional[int] = None) -> List[Dict[str, Any]]:
    results, errors, has_errors = run_parallel('get_item_properties',
                                               [get_item for _ in item_ids],
                                               [(session_info, guid) for guid in item_ids],
                                               max_parallelism=max_parallelism, executor=executor, timeout=timeout)

    return results  # type: ignore


def create_alert_conditions(client, name_suffix: str,
                            formula: str, formula_bindings: Dict[str, Dict[str, Any]],
                            signals: List[Dict], dry_run: bool):
    inputs = []
    for s in signals:
        bindings = formula_bindings[s['name']]
        workbook_url = f"{MONITORS_URL}workbook/{s['scopedTo']}"
        bound_formula = formula.replace(
            '{customer}', s['customer']).replace(
            '{domain}', s['domain']).replace(
            '{server}', s['server']).replace(
            '{threshold}', bindings['threshold']).replace(
            '{min_duration}', bindings['min_duration']).replace(
            '{ignore_gaps}', bindings['ignore_gaps']).replace(
            '{workbook_url}', workbook_url)
        inputs.append(
            ConditionUpdateInputV1(
                name=f"{s['name']} {name_suffix}", scoped_to=s['scopedTo'],
                data_id=f"{s['dataID'].replace('{Signal}', '{Condition}')} {name_suffix}",
                datasource_class=s['datasourceClass'],
                datasource_id=s['datasourceID'],
                formula=bound_formula,
                parameters=[f"s={s['id']}"],
                replace_capsule_properties=True,
                additional_properties=[ScalarPropertyV1('Cache Enabled', None, True),
                                       ScalarPropertyV1('Customer', None, s['customer']),
                                       ScalarPropertyV1('Domain', None, s['domain']),
                                       ScalarPropertyV1('Server', None, s['server']),
                                       ScalarPropertyV1('Workbook', None, workbook_url)]))

    outputs = []
    if not dry_run and len(inputs) > 0:
        response: Dict = sdk.ConditionsApi(client).put_conditions(
            body=ConditionBatchInputV1(conditions=inputs), _response_type='json')  # type: ignore

        for output in response['itemUpdates']:
            if 'item' in output:
                outputs.append(output)
            else:
                print(output['errorMessage'])

    return outputs


def update_condition_monitor(client, conditions: List[Dict]):
    response: Dict = sdk.ConditionMonitorsApi(client).get_condition_monitors(
        name_search='SaaS Application Alerts', limit=1000000, _response_type='json')  # type: ignore

    outputs = []
    for item in response['conditionMonitorItems']:
        condition_ids = {guid.upper() for guid in item['conditionIds']}
        new_condition_ids = {c['item']['id'].upper() for c in conditions}
        if len(condition_ids.symmetric_difference(new_condition_ids)) > 0:
            input = ConditionMonitorInputV1(
                archived=item['isArchived'],
                condition_ids=list(new_condition_ids),
                cron_schedule=item['cronSchedule'],
                # description=item['description'],
                enabled=item['enabled'],
                name=item['name'],
                query_range_look_ahead=item['queryRangeLookAhead'])

            output: Dict = sdk.ConditionMonitorsApi(client).update_condition_monitor(
                id=item['id'], body=input, _response_type='json')  # type: ignore
            outputs.append(output)


def get_alert_signals(session_info: Tuple[str, str, str], delete_existing: bool = False) -> List[Dict]:
    client = get_api_client(session_info)
    if delete_existing:
        response: Dict = sdk.ItemsApi(client).search_items(
            filters=['Name==Cpu.System.UsagePercent.Gauge Alert && Datasource ID==Seeq Monitoring Script',
                     'Name==Processes.Appserver.ServerLoadPercent.Gauge Alert && Datasource ID==Seeq Monitoring '
                     'Script',
                     'Name==Threads.HttpServer.LongRunning.Queued.Timer.p75 Alert && Datasource ID==Seeq Monitoring '
                     'Script'],
            types=['CalculatedCondition'],
            limit=1000000, _response_type='json')  # type: ignore
        delete_items(client, response['items'], True)

    response: Dict = sdk.ItemsApi(client).search_items(
        filters=['Name==Cpu.System.UsagePercent.Gauge && Datasource ID==Seeq Monitoring Script',
                 'Name==Processes.Appserver.ServerLoadPercent.Gauge && Datasource ID==Seeq Monitoring Script',
                 'Name==Threads.HttpServer.LongRunning.Queued.Timer.p75 && Datasource ID==Seeq Monitoring Script'],
        types=['CalculatedSignal'],
        limit=1000000, _response_type='json')  # type: ignore
    items = {i['id']: i for i in response['items'] if not i['name'].endswith(' Alert')}

    signals = get_item_properties(session_info, list(items.keys()))
    return [s for s in [assign_properties(s, s['properties']) for s in signals] if 'customer' in s]


def main(url: str, user: str, access_key: str, password: str, delete_existing: bool):
    spy.options.allow_version_mismatch = True
    spy.login(url=url, access_key=access_key, username=user, password=password, ignore_ssl_errors=True, quiet=True)
    session_info = get_session_info(spy.session)
    client = get_api_client(session_info)
    signals = get_alert_signals(session_info, delete_existing)

    conditions = create_alert_conditions(client, 'Alert', formula, formula_bindings, signals, False)

    update_condition_monitor(client, conditions)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-u", "--user", required=False, help="User name to log in with")
    parser.add_argument("-p", "--password", required=False, help="Password name to log in with")
    parser.add_argument("-k", "--access_key", required=False, help="Access key to log in with")
    parser.add_argument("-d", "--delete_existing", required=False, default=False,
                        help="Deletes existing alert signals when true")

    args = parser.parse_args()
    main(MONITORS_URL, args.user, args.access_key, args.password, args.delete_existing)
